from ..errors import RemouladeError


class StateError(RemouladeError):
    """Base class for State Errors"""


class InvalidStateError(StateError):
    """Raised when you try to declare an state
    that is not defined.
    """



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 WIREMIND SAS <dev@wiremind.fr>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from collections import namedtuple
from typing import Any, Optional

from .broker import get_broker


class Result(namedtuple("Result", ("message_id",))):
    """Encapsulates metadata needed to retrieve the result of a message

    Parameters:
      message_id(str): The id of the message sent to the broker.
    """

    def __new__(cls, *, message_id: Optional[str] = None) -> "Result":
        return super().__new__(cls, message_id=message_id)

    def asdict(self):
        return self._asdict()

    def get(
        self, *, block: bool = False, timeout: Optional[int] = None, raise_on_error: bool = True, forget: bool = False
    ) -> Any:
        """Get the result associated with a message_id from a result backend.

        Parameters:
          block(bool): Whether or not to block while waiting for a
            result.
          timeout(int): The maximum amount of time, in ms, to block
            while waiting for a result.
          raise_on_error(bool): raise an error if the result stored in
            an error
          forget(bool): if true the result is discarded from the result
            backend

        Raises:
          RuntimeError: If there is no result backend on the default
            broker.
          ResultMissing: When block is False and the result isn't set.
          ResultTimeout: When waiting for a result times out.
          ErrorStored: When the result is an error and raise_on_error is True

        Returns:
          object: The result.
        """
        broker = get_broker()
        backend = broker.get_result_backend()

        return backend.get_result(
            self.message_id, block=block, timeout=timeout, forget=forget, raise_on_error=raise_on_error
        )

    def completed(self) -> bool:
        """Returns True when the job has been completed (error or result).

        Raises:
          RuntimeError: If your broker doesn't have a result backend
            set up.
        """
        broker = get_broker()
        backend = broker.get_result_backend()
        return backend.get_status([self.message_id]) == 1



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import platform

from .age_limit import AgeLimit
from .callbacks import Callbacks
from .current_message import CurrentMessage
from .logging_metadata import LoggingMetadata
from .max_memory import MaxMemory
from .middleware import Middleware, MiddlewareError, SkipMessage
from .pipelines import Pipelines
from .retries import Retries
from .shutdown import Shutdown, ShutdownNotifications
from .threading import Interrupt, raise_thread_exception
from .time_limit import TimeLimit, TimeLimitExceeded

CURRENT_OS = platform.system()

if CURRENT_OS != "Windows":
    from .prometheus import Prometheus  # noqa: F401


__all__ = [
    # Basics
    "Middleware",
    "MiddlewareError",
    "SkipMessage",
    # Threading
    "Interrupt",
    "raise_thread_exception",
    # Middlewares
    "AgeLimit",
    "Callbacks",
    "CurrentMessage",
    "LoggingMetadata",
    "Pipelines",
    "Retries",
    "Shutdown",
    "ShutdownNotifications",
    "TimeLimit",
    "TimeLimitExceeded",
    "MaxMemory",
]

if CURRENT_OS != "Windows":
    __all__.append("Prometheus")


#: The list of middleware that are enabled by default.
default_middleware = [AgeLimit, TimeLimit, ShutdownNotifications, Callbacks, Pipelines, Retries, CurrentMessage]



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Callable, List

import redis

from ..backend import RateLimiterBackend


class RedisBackend(RateLimiterBackend):
    """A rate limiter backend for Redis_.

    Parameters:
      client(Redis): An optional client.  If this is passed,
        then all other parameters are ignored.
      url(str): An optional connection URL.  If both a URL and
        connection paramters are provided, the URL is used.
      **parameters(dict): Connection parameters are passed directly
        to :class:`redis.Redis`.

    .. _redis: https://redis.io
    """

    def __init__(self, *, client=None, url=None, **parameters):
        if url is not None:
            parameters["connection_pool"] = redis.ConnectionPool.from_url(url)

        self.client = client or redis.Redis(**parameters)

    def add(self, key: str, value: int, ttl: int) -> bool:
        return bool(self.client.set(key, value, px=ttl, nx=True))

    def incr(self, key: str, amount: int, maximum: int, ttl: int) -> bool:
        with self.client.pipeline() as pipe:
            while True:
                try:
                    pipe.watch(key)
                    value = int(pipe.get(key) or b"0")  # type: ignore
                    value += amount
                    if value > maximum:
                        return False

                    pipe.multi()
                    pipe.set(key, value, px=ttl)
                    pipe.execute()
                    return True
                except redis.WatchError:
                    continue

    def decr(self, key: str, amount: int, minimum: int, ttl: int) -> bool:
        with self.client.pipeline() as pipe:
            while True:
                try:
                    pipe.watch(key)
                    value = int(pipe.get(key) or b"0")  # type: ignore
                    value -= amount
                    if value < minimum:
                        return False

                    pipe.multi()
                    pipe.set(key, value, px=ttl)
                    pipe.execute()
                    return True
                except redis.WatchError:
                    continue

    def incr_and_sum(self, key: str, keys: Callable[[], List[str]], amount: int, maximum: int, ttl: int) -> bool:
        with self.client.pipeline() as pipe:
            while True:
                try:
                    # TODO: Drop non-callable keys in Remoulade v2.
                    key_list = keys() if callable(keys) else keys
                    pipe.watch(key, *key_list)
                    value = int(pipe.get(key) or b"0")  # type: ignore
                    value += amount
                    if value > maximum:
                        return False

                    # Fetch keys again to account for net/server latency.
                    values = pipe.mget(keys() if callable(keys) else keys)
                    total = amount + sum(int(n) for n in values if n)  # type: ignore
                    if total > maximum:
                        return False

                    pipe.multi()
                    pipe.set(key, value, px=ttl)
                    pipe.execute()
                    return True
                except redis.WatchError:
                    continue



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 WIREMIND SAS <dev@wiremind.fr>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import time
from typing import Iterable, Optional

import redis

from ..backend import CancelBackend


class RedisBackend(CancelBackend):
    """A cancel backend for Redis_.

    It uses a sorted set with the message_id as member and the timestamp of the addition as a score.
    We can check if message has been canceled if it belongs to the set.
    And on each message cancel we delete all the cancellations that are older than cancellation_ttl (ZREMRANGEBYSCORE).
    This prevents unlimited set growth.

    Parameters:
      cancellation_ttl(int): The minimal amount of seconds message cancellations
        should be kept in the backend.
      key(str): A string to serve as key for the sorted set.
      client(Redis): An optional client.  If this is passed,
        then all other parameters are ignored.
      url(str): An optional connection URL.  If both a URL and
        connection parameters are provided, the URL is used.
      **parameters(dict): Connection parameters are passed directly
        to :class:`redis.Redis`.

    .. _redis: https://redis.io
    """

    def __init__(
        self,
        *,
        cancellation_ttl: Optional[int] = None,
        key: str = "remoulade-cancellations",
        client: Optional[redis.Redis] = None,
        url: Optional[str] = None,
        **parameters,
    ) -> None:
        super().__init__(cancellation_ttl=cancellation_ttl)

        if url:
            parameters["connection_pool"] = redis.ConnectionPool.from_url(url)

        self.key = key
        self.client = client or redis.Redis(**parameters)

    def is_canceled(self, message_id: str, group_id: str) -> bool:
        with self.client.pipeline() as pipe:
            [pipe.zscore(self.key, key) for key in [message_id, group_id] if key]
            results = pipe.execute()
        return any(result is not None for result in results)

    def cancel(self, message_ids: Iterable[str]) -> None:
        timestamp = time.time()
        with self.client.pipeline() as pipe:
            pipe.zadd(self.key, {message_id: timestamp for message_id in message_ids})
            pipe.zremrangebyscore(self.key, "-inf", timestamp - self.cancellation_ttl)
            pipe.execute()



from .backend import State, StateBackend, StateStatusesEnum
from .errors import InvalidStateError
from .middleware import MessageState

__all__ = ["State", "StateBackend", "MessageState", "StateStatusesEnum", "InvalidStateError"]



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .middleware import Middleware


class Callbacks(Middleware):
    """Middleware that lets you chain success and failure callbacks
    onto Actors.

    Parameters:
      on_failure(str): The name of an actor to send a message to on
        failure.
      on_success(str): The name of an actor to send a message to on
        success.
    """

    @property
    def actor_options(self):
        return {
            "on_failure",
            "on_success",
        }

    def after_process_message(self, broker, message, *, result=None, exception=None):
        if exception is None:
            actor_name = self.get_option("on_success", broker=broker, message=message)
            if actor_name:
                actor = broker.get_actor(actor_name)
                actor.send(message.asdict(), result)

        else:
            actor_name = self.get_option("on_failure", broker=broker, message=message)
            if actor_name:
                actor = broker.get_actor(actor_name)
                actor.send(message.asdict(), {"type": type(exception).__name__, "message": str(exception)})

    def update_options_before_create_message(self, options, broker, actor_name):
        from ..actor import Actor

        for option_name in ["on_failure", "on_success"]:
            callback = options.get(option_name)
            if isinstance(callback, Actor):
                options[option_name] = callback.actor_name

            elif callback is not None and not isinstance(callback, str):
                raise TypeError(f"{option_name} must be an Actor or a string, got {type(callback)} instead")

        return options



from typing import Callable, Dict, Optional

from ..logging import get_logger
from .middleware import Middleware


class LoggingMetadata(Middleware):
    """Middleware that lets you add logging_metadata to messages.

    Parameters:
      logging_metadata(dict): The metadata to add to messages
      logging_metadata_getter(Callable[[], dict]): A callback that returns
    """

    def __init__(
        self, *, logging_metadata: Optional[Dict] = None, logging_metadata_getter: Optional[Callable[[], Dict]] = None
    ):
        self.logger = get_logger(__name__, type(self))
        self.logging_metadata = logging_metadata
        self.logging_metadata_getter = logging_metadata_getter

    @property
    def actor_options(self):
        return {
            "logging_metadata",
            "logging_metadata_getter",
        }

    def merge_metadata(self, total_logging_metadata, logging_metadata, logging_metadata_getter):
        if logging_metadata is not None:
            total_logging_metadata = {**total_logging_metadata, **logging_metadata}
        if logging_metadata_getter is not None:
            if callable(logging_metadata_getter):
                total_logging_metadata = {**total_logging_metadata, **logging_metadata_getter()}
            else:
                raise TypeError("logging_metadata_getter must be callable.")

        return total_logging_metadata

    def update_options_before_create_message(self, options, broker, actor_name):

        total_logging_metadata: dict = {}

        # getting logging_metadata at middleware level
        logging_metadata = self.logging_metadata
        logging_metadata_getter = self.logging_metadata_getter
        total_logging_metadata = self.merge_metadata(total_logging_metadata, logging_metadata, logging_metadata_getter)

        # getting logging_metadata at actor level
        actor_options = broker.get_actor(actor_name).options
        logging_metadata = actor_options.get("logging_metadata", None)
        logging_metadata_getter = actor_options.get("logging_metadata_getter", None)
        total_logging_metadata = self.merge_metadata(total_logging_metadata, logging_metadata, logging_metadata_getter)

        # getting logging_metadata at message level
        logging_metadata = options.get("logging_metadata", None)
        logging_metadata_getter = options.get("logging_metadata_getter", None)
        total_logging_metadata = self.merge_metadata(total_logging_metadata, logging_metadata, logging_metadata_getter)
        if logging_metadata_getter is not None:
            del options["logging_metadata_getter"]

        if total_logging_metadata != {}:
            options["logging_metadata"] = total_logging_metadata

        for name in ["message_id", "input"]:
            if name in total_logging_metadata:
                self.logger.error(f"'{name}' cannot be used as a logging_metadata key. It will be overwritten.")

        return options



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time
from threading import Lock
from typing import Callable, Dict, List, Optional, Tuple

from ..backend import RateLimiterBackend


class StubBackend(RateLimiterBackend):
    """An in-memory rate limiter backend.  For use in unit tests."""

    def __init__(self):
        self.mutex = Lock()
        self.db: Dict[str, Tuple[int, float]] = {}

    def add(self, key: str, value: int, ttl: int) -> bool:
        with self.mutex:
            res = self._get(key)
            if res is not None:
                return False

            return self._put(key, value, ttl)

    def incr(self, key: str, amount: int, maximum: int, ttl: int) -> bool:
        with self.mutex:
            value = self._get(key, default=0) + amount
            if value > maximum:
                return False

            return self._put(key, value, ttl)

    def decr(self, key: str, amount: int, minimum: int, ttl: int) -> bool:
        with self.mutex:
            value = self._get(key, default=0) - amount
            if value < minimum:
                return False

            return self._put(key, value, ttl)

    def incr_and_sum(self, key: str, keys: Callable[[], List[str]], amount: int, maximum: int, ttl: int) -> bool:
        self.add(key, 0, ttl)
        with self.mutex:
            value = self._get(key, default=0) + amount
            if value > maximum:
                return False

            # TODO: Drop non-callable keys in Remoulade v2.
            key_list = keys() if callable(keys) else keys
            values = sum(self._get(k, default=0) for k in key_list)
            total = amount + values
            if total > maximum:
                return False

            return self._put(key, value, ttl)

    def _get(self, key: str, *, default: Optional[int] = None) -> Optional[int]:
        value, expiration = self.db.get(key, (None, None))
        if expiration and time.monotonic() < expiration:
            return value
        return default

    def _put(self, key: str, value: int, ttl: int) -> bool:
        self.db[key] = (value, time.monotonic() + ttl / 1000)
        return True



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Set

from ..logging import get_logger
from ..middleware import Middleware, Retries
from .backend import BackendResult
from .errors import ParentFailed

#: The maximum amount of milliseconds results are allowed to exist in
#: the backend.
DEFAULT_RESULT_TTL = 600000


class Results(Middleware):
    """Middleware that automatically stores actor results.

    Example:

      >>> from remoulade.results import Results
      >>> from remoulade.results.backends import RedisBackend
      >>> backend = RedisBackend()
      >>> broker.add_middleware(Results(backend=backend))

      >>> @remoulade.actor(store_results=True)
      ... def add(x, y):
      ...   return x + y

      >>> broker.declare_actor(add)
      >>> message = add.send(1, 2)
      >>> message.result.get()
      3

    Parameters:
      backend(ResultBackend): The result backend to use when storing
        results.
      store_results(bool): Whether or not actor results should be
        stored.  Defaults to False and can be set on a per-actor
        basis.
      result_ttl(int): The maximum number of milliseconds results are
        allowed to exist in the backend.  Defaults to 10 minutes and
        can be set on a per-actor basis.
    """

    # Warning: Results need to be before Retries to work
    default_before = Retries

    def __init__(self, *, backend=None, store_results=False, result_ttl=None):
        self.logger = get_logger(__name__, type(self))
        self.backend = backend
        self.store_results = store_results
        self.result_ttl = result_ttl or DEFAULT_RESULT_TTL

    @property
    def actor_options(self):
        return {"store_results", "result_ttl"}

    def after_process_message(self, broker, message, *, result=None, exception=None):
        store_results = self.get_option("store_results", broker=broker, message=message)
        result_ttl = self.get_option("result_ttl", broker=broker, message=message)
        message_failed = getattr(message, "failed", False)

        results = []
        if store_results:
            if exception is None:
                results.append((message.message_id, BackendResult(result=result, error=None)))
            elif message_failed:
                error_str = self._serialize_exception(exception)
                results.append((message.message_id, BackendResult(result=None, error=error_str)))

        # even if the actor do not have store_results, we need to invalidate the messages in the pipeline that has it
        if message_failed:
            error_str = self._serialize_exception(exception)
            exception = ParentFailed("%s failed because of %s" % (message, error_str))
            children_result = BackendResult(result=None, error=self._serialize_exception(exception))

            for message_id in self._get_children_message_ids(
                broker, self.get_option("pipe_target", broker=broker, message=message)
            ):
                results.append((message_id, children_result))

        if results:
            message_ids, results = zip(*results)
            with self.backend.retry(broker, message, self.logger):
                self.backend.store_results(message_ids, results, result_ttl)

    @staticmethod
    def _serialize_exception(exception):
        try:
            return repr(exception)
        except Exception as e:  # noqa
            return "Exception could not be serialized"

    def _get_children_message_ids(self, broker, pipe_target):
        """Get the ids of all the following messages in the pipeline which have store_results"""
        from ..message import Message

        message_ids: Set[str] = set()

        if isinstance(pipe_target, list):
            for message_data in pipe_target:
                message_ids |= self._get_children_message_ids(broker, message_data)
        elif pipe_target:
            message = Message(**pipe_target)
            if self.get_option("store_results", broker=broker, message=message):
                message_ids.add(message.message_id)

            message_ids |= self._get_children_message_ids(
                broker, self.get_option("pipe_target", broker=broker, message=message)
            )

        return message_ids

    def after_enqueue_pipe_target(self, _, group_info):
        """After a pipe target has been enqueued, we need to forget the result of the group (if it's a group)"""
        if group_info:
            message_ids = self.backend.get_group_message_ids(group_info.group_id)
            self.backend.forget_results(message_ids, self.result_ttl)
            self.backend.delete_group_message_ids(group_info.group_id)

    def before_build_group_pipeline(self, _, group_id, message_ids):
        self.backend.set_group_message_ids(group_id, message_ids, self.result_ttl)



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .actor import Actor, actor
from .broker import Broker, Consumer, MessageProxy, change_broker, declare_actors, get_broker, set_broker
from .collection_results import CollectionResults
from .composition import group, pipeline
from .encoder import Encoder, JSONEncoder, PickleEncoder
from .errors import (
    ActorNotFound,
    BrokerError,
    ChannelPoolTimeout,
    ConnectionClosed,
    ConnectionError,
    ConnectionFailed,
    NoResultBackend,
    QueueJoinTimeout,
    QueueNotFound,
    RateLimitExceeded,
    RemouladeError,
)
from .generic import GenericActor
from .logging import get_logger
from .message import Message, get_encoder, set_encoder
from .middleware import Middleware
from .result import Result
from .utils import get_scheduler, set_scheduler
from .worker import Worker

__all__ = [
    # Actors
    "Actor",
    "GenericActor",
    "actor",
    # Brokers
    "Broker",
    "Consumer",
    "MessageProxy",
    "get_broker",
    "set_broker",
    "change_broker",
    "declare_actors",
    # Composition
    "group",
    "pipeline",
    "CollectionResults",
    # Encoding
    "Encoder",
    "JSONEncoder",
    "PickleEncoder",
    # Errors
    "RemouladeError",
    "BrokerError",
    "ActorNotFound",
    "QueueNotFound",
    "QueueJoinTimeout",
    "ConnectionError",
    "ConnectionClosed",
    "ConnectionFailed",
    "RateLimitExceeded",
    "NoResultBackend",
    "ChannelPoolTimeout",
    # Logging
    "get_logger",
    # Messages
    "Message",
    "get_encoder",
    "set_encoder",
    "Result",
    # Middlware
    "Middleware",
    # Workers
    "Worker",
    # Scheduler
    "get_scheduler",
    "set_scheduler",
]

__version__ = "0.34.2"



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from collections import namedtuple
from typing import TYPE_CHECKING, Iterable, List, Optional, Union, cast

from typing_extensions import TypedDict

from .broker import get_broker
from .collection_results import CollectionResults
from .common import flatten, generate_unique_id

if TYPE_CHECKING:
    from .message import Message  # noqa
    from .result import Result


class GroupInfoDict(TypedDict):
    group_id: str
    children_count: int
    cancel_on_error: bool


class GroupInfo(namedtuple("GroupInfo", ("group_id", "children_count", "cancel_on_error"))):
    """Encapsulates metadata about a group being sent to multiple actors.

    Parameters:
      group_id(str): The id of the group
      children_count(int)
      cancel_on_error(bool)
    """

    def __new__(cls, *, group_id: str, children_count: int, cancel_on_error: bool):
        return super().__new__(cls, group_id, children_count, cancel_on_error)

    def asdict(self) -> GroupInfoDict:
        return cast(GroupInfoDict, self._asdict())


class pipeline:
    """Chain actors together, passing the result of one actor to the
    next one in line.

    Parameters:
      children(Iterable[Message|pipeline|group]): A sequence of messages or
        pipelines or groups.  Child pipelines are flattened into the resulting
        pipeline.
      broker(Broker): The broker to run the pipeline on.  Defaults to
        the current global broker.

    Attributes:
        children(List[Message|group]) The sequence of messages or groups to execute as a pipeline
    """

    def __init__(self, children: "Iterable[Union[Message, pipeline, group]]"):
        self.broker = get_broker()

        self.children: List[Union[Message, group]] = []
        for child in children:
            if isinstance(child, pipeline):
                self.children += child.children
            elif isinstance(child, group):
                self.children.append(child)
            else:
                self.children.append(child.copy())

    def build(self, *, last_options=None):
        """Build the pipeline, return the first message to be enqueued or integrated in another pipeline

        Build the pipeline by starting at the end. We build a message with all it's options in one step and
        we serialize it (asdict) as the previous message pipe_target in the next step.

        We need to know what is the options (pipe_target) of the pipeline before building it because we cannot
        edit the pipeline after it has been built.

        Parameters:
            last_options(dict): options to be assigned to the last actor of the pipeline (ex: pipe_target)

        Returns:
            the first message of the pipeline
        """
        next_child = None
        for child in reversed(self.children):
            if next_child:
                options = {"pipe_target": [m.asdict() for m in next_child]}
            else:
                options = last_options or {}

            if isinstance(child, group):
                next_child = child.build(options)
            else:
                next_child = [child.build(options)]

        return next_child

    def __len__(self):
        """Returns the length of the pipeline."""
        return len(self.children)

    def __or__(self, other: "Union[Message, group]"):
        """Returns a new pipeline with "other" added to the end."""
        return type(self)(self.children + [other])

    def __str__(self):  # pragma: no cover
        return "pipeline([%s])" % ", ".join(str(m) for m in self.children)

    @property
    def message_ids(self):
        for child in self.children:
            if isinstance(child, group):
                yield list(child.message_ids)
            else:
                yield child.message_id

    def run(self, *, delay: Optional[int] = None) -> "pipeline":
        """Run this pipeline.

        Parameters:
          delay(int): The minimum amount of time, in milliseconds, the
            pipeline should be delayed by.

        Returns:
          pipeline: Itself.
        """
        first = self.build()
        if isinstance(first, list):
            for message in first:
                self.broker.enqueue(message, delay=delay)
        else:
            self.broker.enqueue(first, delay=delay)
        return self

    @property
    def results(self) -> CollectionResults:
        """CollectionResults created from this pipeline, used for result related methods"""
        results: List[Union[Result, CollectionResults]] = []
        for element in self.children:
            results += [element.results if isinstance(element, group) else element.result]
        return CollectionResults(results)

    @property
    def result(self):
        """Result of the last message/group of the pipeline"""
        last_child = self.children[-1]
        return last_child.results if isinstance(last_child, group) else last_child.result

    def cancel(self) -> None:
        """Mark all the children as cancelled"""
        broker = get_broker()
        backend = broker.get_cancel_backend()
        backend.cancel(list(flatten(self.message_ids)))


class group:
    """Run a group of actors in parallel.

    Parameters:
      children(Iterable[Message|pipeline]): A sequence of messages or pipelines.
      cancel_on_error(boolean): True if you want to cancel all messages of a group if on of
        the actor fails, this is only possible with a Cancel middleware.

    Attributes:
        children(List[Message|pipeline]) The sequence to execute as a group

    Raise:
        NoCancelBackend: if no cancel middleware is set
    """

    def __init__(
        self,
        children: "Iterable[Union[Message, pipeline]]",
        *,
        group_id: Optional[str] = None,
        cancel_on_error: bool = False,
    ) -> None:
        self.children: "List[Union[Message, pipeline]]" = []
        for child in children:
            if isinstance(child, group):
                raise ValueError("Groups of groups are not supported")
            self.children.append(child)

        self.broker = get_broker()
        self.group_id: str = generate_unique_id() if group_id is None else group_id
        self.cancel_on_error = cancel_on_error
        if cancel_on_error:
            self.broker.get_cancel_backend()

    def __or__(self, other: "Union[Message, group, pipeline]") -> pipeline:
        """Combine this group into a pipeline with "other"."""
        return pipeline([self, other])

    def __len__(self) -> int:
        """Returns the size of the group."""
        return len(self.children)

    def __str__(self) -> str:  # pragma: no cover
        return f"group({', '.join(str(child) for child in self.children)})"

    def build(self, options=None) -> "List[Message]":
        """Build group for pipeline"""
        if options is None:
            options = {}
        else:
            self.broker.emit_before("build_group_pipeline", group_id=self.group_id, message_ids=list(self.message_ids))

        options = {"group_info": self.info.asdict(), **options}
        messages: "List[Message]" = []
        for group_child in self.children:
            if isinstance(group_child, pipeline):
                messages += group_child.build(last_options=options)
            else:
                messages += [group_child.build(options)]
        return messages

    @property
    def info(self) -> GroupInfo:
        """Info used for group completion and cancel"""
        return GroupInfo(
            group_id=self.group_id, children_count=len(self.children), cancel_on_error=self.cancel_on_error
        )

    @property
    def message_ids(self):
        for child in self.children:
            if isinstance(child, pipeline):
                yield list(child.message_ids)
            else:
                yield child.message_id

    def run(self, *, delay: Optional[int] = None) -> "group":
        """Run the actors in this group.

        Parameters:
          delay(int): The minimum amount of time, in milliseconds,
            each message in the group should be delayed by.
        """
        for message in self.build():
            self.broker.enqueue(message, delay=delay)

        return self

    @property
    def results(self) -> CollectionResults:
        """CollectionResults created from this group, used for result related methods"""
        return CollectionResults(children=[child.result for child in self.children])

    def cancel(self) -> None:
        """Mark all the children as cancelled"""
        broker = get_broker()
        backend = broker.get_cancel_backend()
        backend.cancel(list(flatten(self.message_ids)))



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 WIREMIND SAS <dev@wiremind.fr>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import TYPE_CHECKING, Optional

from .errors import NoScheduler

if TYPE_CHECKING:
    from .scheduler import Scheduler

global_scheduler: "Optional[Scheduler]" = None


def get_scheduler() -> "Scheduler":
    global global_scheduler
    if global_scheduler is None:
        raise NoScheduler("Scheduler not found, are you sure you called set_scheduler(scheduler) ?")
    return global_scheduler


def set_scheduler(scheduler: "Scheduler") -> None:
    global global_scheduler
    global_scheduler = scheduler



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 WIREMIND SAS <dev@wiremind.fr>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ..composition import group


def reduce(messages, merge_actor, cancel_on_error=False, size=2, merge_kwargs=None):
    """Recursively merge messages

    Parameters:
      messages(Iterator[Message|pipeline]): A sequence of messages or pipelines that needs to be merged.
      merge_actor(Actor): The actor that will be responsible for the merge of two messages.
      cancel_on_error(boolean): True if you want to cancel all messages of a group if one of
        the actor fails, this is only possible with a Cancel middleware.
      size(int): Number of messages that are reduced at once.
      merge_kwargs: kwargs to be passed to each merge message.

    Returns:
      Message|pipeline: a message or a pipeline that will return the reduced result of all the given messages.

    Raise:
        NoCancelBackend: if no cancel middleware is set
    """
    if merge_kwargs is None:
        merge_kwargs = {}
    messages = list(messages)
    while len(messages) > 1:
        reduced_messages = []
        for i in range(0, len(messages), size):
            if i == len(messages) - (size - 1):
                reduced_messages.append(messages[i])
            else:
                grouped_message = group(messages[i : i + size], cancel_on_error=cancel_on_error)
                reduced_messages.append(grouped_message | merge_actor.message(**merge_kwargs))
        messages = reduced_messages

    return messages[0]



from datetime import datetime, timezone

from ..middleware import Middleware
from .backend import State, StateStatusesEnum


class MessageState(Middleware):

    """Middleware use to storage and update the state
    of the messages.
    Parameters
    state_ttl(int):Time(seconds) that the state will be storage
                    in the database
    """

    def __init__(self, backend, state_ttl=3600):
        self.backend = backend
        self.state_ttl = state_ttl

    def save(self, message, status, priority=None, **kwargs):
        if self.state_ttl is None or self.state_ttl <= 0:
            return
        args = message.args
        options = message.options
        kwargs_state = message.kwargs
        message_id = message.message_id
        actor_name = message.actor_name
        self.backend.set_state(
            State(
                message_id,
                status,
                actor_name=actor_name,
                args=args,
                priority=priority,
                kwargs=kwargs_state,
                options=options,
                **kwargs,
            ),
            self.state_ttl,
        )

    def _get_current_time(self):
        return datetime.now(timezone.utc)

    def after_enqueue(self, broker, message, delay):
        priority = broker.get_actor(message.actor_name).priority
        group_id = self.get_option("group_info", broker=broker, message=message, default={}).get("group_id")
        self.save(
            message,
            status=StateStatusesEnum.Pending,
            enqueued_datetime=self._get_current_time(),
            priority=priority,
            group_id=group_id,
        )

    def after_skip_message(self, broker, message):
        self.save(message, status=StateStatusesEnum.Skipped)

    def after_message_canceled(self, broker, message):
        self.save(message, status=StateStatusesEnum.Canceled)

    def after_process_message(self, broker, message, *, result=None, exception=None):
        self.save(
            message,
            status=StateStatusesEnum.Success if exception is None else StateStatusesEnum.Failure,
            end_datetime=self._get_current_time(),
        )

    def before_process_message(self, broker, message):
        self.save(message, status=StateStatusesEnum.Started, started_datetime=self._get_current_time())



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 WIREMIND SAS <dev@wiremind.fr>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ..logging import get_logger
from ..middleware import Middleware
from .errors import MessageCanceled


class Cancel(Middleware):
    """Middleware that check if a message has been canceled before processing it
    If the message has been canceled raise a MessageCanceled to prevent message processing.

    Example:

      >>> from remoulade.cancel import Cancel
      >>> from remoulade.cancel.backends import RedisBackend
      >>> backend = RedisBackend()
      >>> broker.add_middleware(Cancel(backend=backend))

      >>> @remoulade.actor(store_results=True)
      ... def add(x, y):
      ...   return x + y

      >>> broker.declare_actor(add)
      >>> message = add.send(1, 2)
      >>> message.cancel()
      3

    Parameters:
      backend(CancelBackend): The cancel backend to use to check
        cancellations.
        Defaults to False and can be set on a per-actor basis.
    """

    def __init__(self, *, backend=None):
        self.logger = get_logger(__name__, type(self))
        self.backend = backend

    def before_process_message(self, broker, message):
        group_id = self.get_option("group_info", broker=broker, message=message, default={}).get("group_id")

        if self.backend.is_canceled(message.message_id, group_id):
            raise MessageCanceled("Message %s has been canceled" % message.message_id)

    def after_process_message(self, broker, message, *, result=None, exception=None):
        """Cancel all the messages in the group if one of the message of the group fail"""
        from ..composition import GroupInfo

        if exception is None:
            return

        group_info = self.get_option("group_info", broker=broker, message=message)
        if not group_info:
            return

        group_info = GroupInfo(**group_info)
        if group_info.cancel_on_error:
            self.backend.cancel([group_info.group_id])



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ..errors import RemouladeError


class ResultError(RemouladeError):
    """Base class for result errors."""


class ResultTimeout(ResultError):
    """Raised when waiting for a result times out."""


class ResultMissing(ResultError):
    """Raised when a result can't be found."""


class ErrorStored(ResultError):
    """Raised when an error is stored in the result backend and raise_on_error is True"""


class ParentFailed(ResultError):
    """Error stored when a parent actor in the pipeline failed"""


class MessageIdsMissing(ResultError):
    """Raised when message_ids linked to a group_id can't be found"""



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import threading
import time
import warnings
from typing import TYPE_CHECKING, Dict, Optional, Set, Tuple

from ..errors import NoCancelBackend
from ..logging import get_logger
from .middleware import Middleware
from .threading import Interrupt, current_platform, raise_thread_exception, supported_platforms

if TYPE_CHECKING:
    from ..broker import Broker
    from ..message import Message


class TimeLimitExceeded(Interrupt):
    """Exception used to interrupt worker threads when actors exceed
    their time limits.
    """


class TimeLimit(Middleware):
    """Middleware that cancels actors that run for too long.
    Currently, this is only available on CPython.

    Note:
      This works by setting an async exception in the worker thread
      that runs the actor.  This means that the exception will only get
      called the next time that thread acquires the GIL.  Concretely,
      this means that this middleware can't cancel system calls.

    Parameters:
      time_limit(int): The maximum number of milliseconds actors may run for.
      interval(int): The interval (in milliseconds) with which to check for actors that have exceeded the limit.
      exit_delay(int): The delay (in milliseconds) after with we stop (SystemExit) to the worker if the exception failed
       to stop the message (ie. system calls). None to disable, disabled by default.
    """

    def __init__(self, *, time_limit: int = 1800000, interval: int = 1000, exit_delay: Optional[int] = None) -> None:
        self.logger = get_logger(__name__, type(self))
        self.time_limit = time_limit
        self.interval = interval
        self.deadlines: "Dict[int, Optional[Tuple[float, bool, Message]]]" = {}
        self.exit_delay = exit_delay
        self.lock = threading.Lock()

    def _handle(self, *_) -> None:
        with self.lock:
            current_time = time.monotonic()
            for thread_id, row in self.deadlines.items():
                if row is None:
                    continue

                deadline, exception_raised, message = row
                if exception_raised and self.exit_delay and current_time >= deadline + self.exit_delay / 1000:
                    self.deadlines[thread_id] = None
                    try:
                        message.cancel()
                        self.logger.critical(
                            "Could not stop message %s execution with TimeLimitExceeded, "
                            "cancel message and exit the worker",
                            message.message_id,
                        )
                        raise SystemExit(1)
                    except NoCancelBackend:
                        # Sending SIGKILL would requeue the message via RabbitMQ
                        self.logger.error(
                            "Could not stop message %s execution with TimeLimitExceeded, "
                            "but do not stop the worker as the message would be requeue",
                            message.message_id,
                        )
                elif not exception_raised and current_time >= deadline:
                    self.logger.warning("Time limit exceeded. Raising exception in worker thread %r.", thread_id)
                    self.deadlines[thread_id] = deadline, True, message
                    raise_thread_exception(thread_id, TimeLimitExceeded)

    def _timer(self):
        while True:
            try:
                self._handle()
            except:  # noqa
                self.logger.exception("Unhandled error while running the time limit handler.", exc_info=True)

            time.sleep(self.interval / 1000)

    @property
    def actor_options(self) -> Set[str]:
        return {"time_limit"}

    def after_process_boot(self, _) -> None:
        if current_platform in supported_platforms:
            thread = threading.Thread(target=self._timer, daemon=True)
            thread.start()
        else:
            msg = f"TimeLimit cannot kill threads on your current platform {current_platform}."
            warnings.warn(msg, category=RuntimeWarning, stacklevel=2)

    def before_process_message(self, broker: "Broker", message: "Message") -> None:
        limit = self.get_option("time_limit", broker=broker, message=message)
        deadline = time.monotonic() + limit / 1000
        with self.lock:
            self.deadlines[threading.get_ident()] = deadline, False, message

    def after_process_message(self, broker: "Broker", message: "Message", *, result=None, exception=None) -> None:
        with self.lock:
            self.deadlines[threading.get_ident()] = None

    after_skip_message = after_process_message
    after_message_canceled = after_process_message



import argparse
import importlib
import json
import sys

from remoulade import get_broker


def parse_arguments():
    parser = argparse.ArgumentParser(
        prog="remoulade-run", description="Runs a remoulade actor", formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument("modules", metavar="module", nargs="*", help="additional python modules to import")
    parser.add_argument("--path", "-P", default=".", nargs="*", type=str, help="the module import path (default: .)")
    parser.add_argument("--actor-name", "-N", type=str, help="The actor to be ran")
    parser.add_argument("--args", "-A", type=json.loads, help="The actor's args")
    parser.add_argument("--kwargs", "-K", type=json.loads, help="The actor's kwargs")
    return parser.parse_args()


def main():

    args = parse_arguments()

    for path in args.path:
        sys.path.insert(0, path)

    for module in args.modules:
        importlib.import_module(module)

    if args.actor_name not in get_broker().actors:
        print("{} is not an available actor".format(args.actor_name))
        return

    _args = args.args or []
    kwargs = args.kwargs or {}

    print(get_broker().actors[args.actor_name](*_args, **kwargs))



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import List

import redis

from ..backend import BackendResult, ForgottenResult, Missing, ResultBackend, ResultMissing, ResultTimeout


class RedisBackend(ResultBackend):
    """A result backend for Redis_.  This is the recommended result
    backend as waiting for a result is resource efficient.

    Parameters:
      namespace(str): A string with which to prefix result keys.
      encoder(Encoder): The encoder to use when storing and retrieving
        result data.  Defaults to :class:`.JSONEncoder`.
      client(Redis): An optional client.  If this is passed,
        then all other parameters are ignored.
      url(str): An optional connection URL.  If both a URL and
        connection paramters are provided, the URL is used.
      **parameters(dict): Connection parameters are passed directly
        to :class:`redis.Redis`.

    .. _redis: https://redis.io
    """

    def __init__(
        self, *, namespace="remoulade-results", encoder=None, client=None, url=None, default_timeout=None, **parameters
    ):
        super().__init__(namespace=namespace, encoder=encoder, default_timeout=default_timeout)

        if url:
            parameters["connection_pool"] = redis.ConnectionPool.from_url(url)

        self.client = client or redis.Redis(**parameters)

    def get_result(self, message_id: str, *, block=False, timeout=None, forget=False, raise_on_error=True):
        """Get a result from the backend.

        Warning:
          Sub-second timeouts are not respected by this backend.

        Parameters:
          message_id(str)
          block(bool): Whether or not to block until a result is set.
          timeout(int): The maximum amount of time, in ms, to wait for
            a result when block is True.  Defaults to 10 seconds.
          forget(bool): Whether or not the result need to be kept.
          raise_on_error(bool): raise an error if the result stored in
            an error

        Raises:
          ResultMissing: When block is False and the result isn't set.
          ResultTimeout: When waiting for a result times out.

        Returns:
          object: The result.
        """
        if timeout is None:
            timeout = self.default_timeout

        message_key = self.build_message_key(message_id)
        timeout = int(timeout / 1000)
        if block and timeout > 0:
            if forget:
                data = self.client.brpoplpush(message_key, message_key, timeout=timeout)
                if data:
                    with self.client.pipeline() as pipe:
                        pipe.lpushx(message_key, self.encoder.encode(ForgottenResult.asdict()))
                        pipe.ltrim(message_key, 0, 0)
                        pipe.execute()
                else:
                    data = None
            else:
                data = self.client.brpoplpush(message_key, message_key, timeout)

            if data is None:
                raise ResultTimeout(message_id)

        else:
            if forget:
                with self.client.pipeline() as pipe:
                    pipe.rpushx(message_key, self.encoder.encode(ForgottenResult.asdict()))
                    pipe.lpop(message_key)
                    data = pipe.execute()[1]
            else:
                data = self.client.rpoplpush(message_key, message_key)

        if data is None:
            if block:
                raise ResultTimeout(message_id)
            else:
                raise ResultMissing(message_id)

        result = BackendResult(**self.encoder.decode(data))
        return self.process_result(result, raise_on_error)

    def _store(self, message_keys, results, ttl):
        with self.client.pipeline() as pipe:
            for (message_key, result) in zip(message_keys, results):
                pipe.delete(message_key)
                pipe.lpush(message_key, self.encoder.encode(result))
                pipe.pexpire(message_key, ttl)
            pipe.execute()

    def _get(self, key, forget=False):
        data = self.client.rpoplpush(key, key)
        if data:
            return self.encoder.decode(data)
        return Missing

    def _delete(self, key):
        self.client.delete(key)

    def increment_group_completion(self, group_id: str, message_id: str) -> int:
        # there should be a ttl
        ttl = 600  # 10 min ?
        group_completion_key = self.build_group_completion_key(group_id)
        with self.client.pipeline() as pipe:
            pipe.sadd(group_completion_key, message_id)
            pipe.expire(group_completion_key, ttl)
            pipe.scard(group_completion_key)
            group_completion = pipe.execute()[2]

        return group_completion

    def get_status(self, message_ids: List[str]) -> int:  # type: ignore
        return self.client.exists(*[self.build_message_key(message_id) for message_id in message_ids])



import time
from typing import Dict

from ..backend import State, StateBackend


class StubBackend(StateBackend):
    """An in-memory state backend.  For use in unit tests.

    Parameters:
      namespace(str): A string with which to prefix result keys.
      encoder(Encoder): The encoder to use when storing and retrieving
        result data.  Defaults to :class:`.JSONEncoder`.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.states = {}  # type: Dict[str, Dict[str, str]]

    def get_state(self, message_id):
        message_key = self._build_message_key(message_id)
        data = self.states.get(message_key)
        state = None
        if data:
            if time.monotonic() > float(data["expiration"]):
                self._delete(message_key)
            else:
                state = State.from_dict(self._decode_dict(data["state"]))
        return state

    def set_state(self, state, ttl=3600):
        message_key = self._build_message_key(state.message_id)
        ttl = ttl + time.monotonic()
        encoded_state = self._encode_dict(state.as_dict())
        if message_key not in self.states.keys():
            payload = {"state": encoded_state, "expiration": ttl}
            self.states[message_key] = payload
        else:
            state = self.states[message_key]["state"]
            for (key, value) in encoded_state.items():
                state[key] = value
            self.states[message_key]["state"] = state
            self.states[message_key]["expiration"] = ttl

    def _delete(self, message_key):
        del self.states[message_key]

    def get_states(self):
        time_now = time.monotonic()
        for message_key in list(self.states.keys()):
            data = self.states[message_key]
            if time_now > float(data["expiration"]):
                self._delete(message_key)
                continue
            state = State.from_dict(self._decode_dict(data["state"]))
            yield state



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import atexit
import importlib
import logging
import os
import signal
import sys
import time

from remoulade import Worker, __version__, get_broker, get_logger

#: The exit codes that the master process returns.
RET_OK = 0  # The process terminated successfully.
RET_KILLED = 1  # The process was killed.
RET_IMPORT = 2  # Module import(s) failed or invalid command line argument.
RET_CONNECT = 3  # Broker connection failed.
RET_PIDFILE = 4  # PID file points to an existing process or cannot be written to.
RET_WORKER_STOP = 5  # Worker was not running anymore

#: The logging format.
logformat = "[%(asctime)s] [%(threadName)s] [%(name)s] [%(levelname)s] %(message)s"

#: The logging verbosity levels.
verbosity = {
    0: logging.INFO,
    1: logging.DEBUG,
}

#: Message printed after the help text.
HELP_EPILOG = """\
examples:
  # Run remoulade workers with actors defined in `./some_module.py`.
  $ remoulade some_module

  # Run remoulade with 1 thread per process.
  $ remoulade --threads 1 some_module

  # Run remoulade with gevent.  Make sure you `pip install gevent` first.
  $ remoulade-gevent --threads 1024 some_module

  # Import extra modules.  Useful when your main module doesn't import
  # all the modules you need.
  $ remoulade some_module some_other_module

  # Listen only to the "foo" and "bar" queues.
  $ remoulade some_module -Q foo bar

  # Write the main process pid to a file.
  $ remoulade some_module --pid-file /tmp/remoulade.pid

  # Write logs to a file.
  $ remoulade some_module --log-file /tmp/remoulade.log
"""


def parse_arguments():
    parser = argparse.ArgumentParser(
        prog="remoulade",
        description="Run remoulade workers.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=HELP_EPILOG,
    )
    parser.add_argument("modules", metavar="module", nargs="*", help="additional python modules to import")
    parser.add_argument(
        "--threads", "-t", default=8, type=int, help="the number of worker threads per process (default: 8)"
    )
    parser.add_argument(
        "--prefetch-multiplier",
        default=2,
        type=int,
        help="""
            the number of messages to prefetch at a time to be multiplied by the number of concurrent processes
            (default:2)
        """,
    )
    parser.add_argument("--path", "-P", default=".", nargs="*", type=str, help="the module import path (default: .)")
    parser.add_argument(
        "--queues", "-Q", nargs="*", type=str, help="listen to a subset of queues (default: all queues)"
    )
    parser.add_argument(
        "--pid-file", type=str, help="write the PID of the master process to a file (default: no pid file)"
    )
    parser.add_argument(
        "--log-file",
        type=argparse.FileType(mode="a", encoding="utf-8"),
        help="write all logs to a file (default: sys.stderr)",
    )
    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument("--verbose", "-v", action="count", default=0, help="turn on verbose log output")
    return parser.parse_args()


def setup_pidfile(filename):
    try:
        pid = os.getpid()
        with open(filename, "r") as pid_file:
            old_pid = int(pid_file.read().strip())
            # This can happen when reloading the process via SIGHUP.
            if old_pid == pid:
                return pid

        try:
            os.kill(old_pid, 0)
            raise RuntimeError("Remoulade is already running with PID %d." % old_pid)
        except OSError:
            try:
                os.remove(filename)
            except FileNotFoundError:
                pass

    except FileNotFoundError:  # pragma: no cover
        pass

    except ValueError:
        # Abort here to avoid overwriting real files.  Eg. someone
        # accidentally specifies a config file as the pid file.
        raise RuntimeError("PID file contains garbage. Aborting.")

    try:
        with open(filename, "w") as pid_file:
            pid_file.write(str(pid))

        # Change permissions to -rw-r--r--.
        os.chmod(filename, 0o644)
        return pid
    except (FileNotFoundError, PermissionError) as e:
        raise RuntimeError("Failed to write PID file %r. %s." % (e.filename, e.strerror))


def remove_pidfile(filename, logger):
    try:
        logger.debug("Removing PID file %r.", filename)
        os.remove(filename)
    except FileNotFoundError:  # pragma: no cover
        logger.debug("Failed to remove PID file. It's gone.")


def setup_logging(args, *, stream=sys.stderr):
    level = verbosity.get(args.verbose, logging.DEBUG)
    logging.basicConfig(level=level, format=logformat, stream=stream)
    return get_logger("remoulade")


def start_worker(args, logger):
    try:
        for module in args.modules:
            importlib.import_module(module)

        broker = get_broker()
        broker.emit_after("process_boot")

        worker = Worker(
            broker, queues=args.queues, worker_threads=args.threads, prefetch_multiplier=args.prefetch_multiplier
        )
        worker.start()
    except ImportError:
        logger.exception("Failed to import module.")
        return os._exit(RET_IMPORT)

    def termhandler(signum, frame):
        nonlocal running  # type: ignore
        if running:
            logger.info("Stopping worker...")
            running = False
        else:
            logger.warning("Killing worker...")
            return os._exit(RET_KILLED)

    logger.info("Worker is ready for action.")
    signal.signal(signal.SIGINT, termhandler)
    signal.signal(signal.SIGTERM, termhandler)
    signal.signal(signal.SIGHUP, termhandler)

    running = True
    ret_code = RET_OK
    while running:
        if worker.consumer_stopped:
            running = False
            ret_code = RET_CONNECT
        if worker.worker_stopped:
            running = False
            ret_code = RET_WORKER_STOP
            logger.info("Worker thread is not running anymore, stopping Worker.")
        else:
            time.sleep(1)

    worker.stop()
    broker.emit_before("process_stop")
    broker.close()
    return ret_code


def main():  # noqa
    args = parse_arguments()
    for path in args.path:
        sys.path.insert(0, path)

    try:
        if args.pid_file:
            setup_pidfile(args.pid_file)
    except RuntimeError as e:
        logger = setup_logging(args, stream=args.log_file or sys.stderr)
        logger.critical(e)
        return RET_PIDFILE

    logger = setup_logging(args, stream=args.log_file or sys.stderr)
    logger.info("Remoulade %r is booting up." % __version__)
    if args.pid_file:
        atexit.register(remove_pidfile, args.pid_file, logger)

    return start_worker(args, logger)


if __name__ == "__main__":
    sys.exit(main())



import os

from setuptools import setup


def rel(*xs):
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), *xs)


with open(rel("README.md")) as f:
    long_description = f.read()


with open(rel("remoulade", "__init__.py"), "r") as f:
    version_marker = "__version__ = "
    for line in f:
        if line.startswith(version_marker):
            _, version = line.split(version_marker)
            version = version.strip().strip('"')
            break
    else:
        raise RuntimeError("Version marker not found.")


dependencies = ["prometheus-client>=0.2", "pytz", "python-dateutil>=2.8.0", "typing-extensions>=3.7"]

extra_dependencies = {
    "rabbitmq": ["amqpstorm>=2.6,<3"],
    "redis": ["redis>=3.5.0,<4.0"],
    "server": ["flask>=1.1,<2", "marshmallow>=3"],
}

extra_dependencies["all"] = list(set(sum(extra_dependencies.values(), [])))
extra_dependencies["dev"] = extra_dependencies["all"] + [
    # Docs
    "alabaster",
    "sphinx==4.1.1",
    "sphinxcontrib-napoleon",
    "sphinxcontrib-versioning",
    "sphinx-copybutton",
    # Linting
    "flake8",
    "flake8-bugbear",
    "flake8-quotes",
    "isort",
    "black==21.7b0",
    "mypy",
    "types-redis",
    "types-python-dateutil",
    # Misc
    "bumpversion",
    "hiredis",
    "twine",
    # Testing
    "pytest",
    "pytest-benchmark[histogram]",
    "pytest-cov",
    "tox",
    "freezegun",
]

setup(
    name="remoulade",
    version=version,
    author="Wiremind",
    author_email="dev@wiremind.fr",
    description="Background Processing for Python 3.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=[
        "remoulade",
        "remoulade.brokers",
        "remoulade.cancel",
        "remoulade.cancel.backends",
        "remoulade.middleware",
        "remoulade.rate_limits",
        "remoulade.rate_limits.backends",
        "remoulade.results",
        "remoulade.results.backends",
        "remoulade.scheduler",
        "remoulade.state",
        "remoulade.state.backends",
        "remoulade.cli",
        "remoulade.helpers",
        "remoulade.api",
    ],
    package_data={"remoulade": ["py.typed"]},
    include_package_data=True,
    install_requires=dependencies,
    python_requires=">=3.6",
    extras_require=extra_dependencies,
    entry_points={
        "console_scripts": [
            "remoulade = remoulade.__main__:main",
            "remoulade-ls = remoulade.cli.remoulade_ls:main",
            "remoulade-run = remoulade.cli.remoulade_run:main",
            "remoulade-scheduler = remoulade.cli.remoulade_scheduler:main",
        ]
    },
    scripts=["bin/remoulade-gevent"],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: System :: Distributed Computing",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
    ],
)



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Type


class MiddlewareError(Exception):
    """Base class for middleware errors."""


class SkipMessage(MiddlewareError):
    """An exception that may be raised by Middleware inside the
    ``before_process_message`` hook in order to skip a message.
    """


class Middleware:
    """Base class for broker middleware.  The default implementations
    for all hooks are no-ops and subclasses may implement whatever
    subset of hooks they like.
    """

    # Used to placed the middleware in the middleware list.
    # Warning: should not be both set and should belong to default middleware
    default_before = None  # type: Type[Middleware]
    default_after = None  # type: Type[Middleware]

    def get_option(self, option_name, *, broker, message, default=None):

        # get option at message level
        option = message.options.get(option_name)
        if option is not None:
            return option

        # it doesn't make sense to get pipe_target and group_info at any other level than message
        if option == "pipe_target" or option == "group_info":
            return None

        # get option at actor level
        actor = broker.get_actor(message.actor_name)
        option = actor.options.get(option_name)
        if option is not None:
            return option

        # get option at middleware level or return default value
        return getattr(self, option_name, default)

    @property
    def actor_options(self):
        """The set of options that may be configured on each actor."""
        return set()

    def before_ack(self, broker, message):
        """Called before a message is acknowledged."""

    def after_ack(self, broker, message):
        """Called after a message has been acknowledged."""

    def before_nack(self, broker, message):
        """Called before a message is rejected."""

    def after_nack(self, broker, message):
        """Called after a message has been rejected."""

    def before_declare_actor(self, broker, actor):
        """Called before an actor is declared."""

    def after_declare_actor(self, broker, actor):
        """Called after an actor has been declared."""

    def before_declare_queue(self, broker, queue_name):
        """Called before a queue is declared."""

    def after_declare_queue(self, broker, queue_name):
        """Called after a queue has been declared."""

    def after_declare_delay_queue(self, broker, queue_name):
        """Called after a delay queue has been declared."""

    def before_enqueue(self, broker, message, delay):
        """Called before a message is enqueued."""

    def after_enqueue(self, broker, message, delay):
        """Called after a message has been enqueued."""

    def before_delay_message(self, broker, message):
        """Called before a message has been delayed in worker memory."""

    def before_process_message(self, broker, message):
        """Called before a message is processed.

        Raises:
          SkipMessage: If the current message should be skipped.  When
            this is raised, ``after_skip_message`` is emitted instead
            of ``after_process_message``.
        """

    def after_process_message(self, broker, message, *, result=None, exception=None):
        """Called after a message has been processed."""

    def after_worker_thread_process_message(self, broker, thread):
        """Called after a worker thread has finished processing a message"""

    def after_skip_message(self, broker, message):
        """Called instead of ``after_process_message`` after a message
        has been skippped.
        """

    def after_message_canceled(self, broker, message):
        """Called instead of ``after_process_message`` after a message
        has been canceled.
        """

    def after_process_boot(self, broker):
        """Called immediately after subprocess start up."""

    def before_process_stop(self, broker):
        """Called before after subprocess stop."""

    def before_worker_boot(self, broker, worker):
        """Called before the worker process starts up."""

    def after_worker_boot(self, broker, worker):
        """Called after the worker process has started up."""

    def before_worker_shutdown(self, broker, worker):
        """Called before the worker process shuts down."""

    def after_worker_shutdown(self, broker, worker):
        """Called after the worker process shuts down."""

    def before_consumer_thread_shutdown(self, broker, thread):
        """Called before a consumer thread shuts down.  This may be
        used to clean up thread-local resources (such as Django
        database connections).

        There is no ``after_consumer_thread_boot``.
        """

    def before_worker_thread_shutdown(self, broker, thread):
        """Called before a worker thread shuts down.  This may be used
        to clean up thread-local resources (such as Django database
        connections).

        There is no ``after_worker_thread_boot``.
        """

    def after_enqueue_pipe_target(self, broker, group_info):
        """Called after the pipe target of a message has been enqueued"""

    def before_build_group_pipeline(self, broker, group_id, message_ids):
        """Called before a group in a group pipeline is enqueued"""

    def update_options_before_create_message(self, options, broker, actor_name):
        """Called when a message is being built.
        The message options is set to this function's return value"""

        return options



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import time
from itertools import chain
from queue import Empty, Queue
from typing import List

from ..broker import Broker, Consumer, MessageProxy
from ..common import current_millis, dq_name, iter_queue, join_queue
from ..errors import QueueNotFound
from ..message import Message


class StubBroker(Broker):
    """A broker that can be used within unit tests.

    Attributes:
      dead_letters(list[Message]): Contains the dead-lettered messages
        for all defined queues.
    """

    def __init__(self, middleware=None):
        super().__init__(middleware)

        self.dead_letters: List[Message] = []

    def consume(self, queue_name, prefetch=1, timeout=100):
        """Create a new consumer for a queue.

        Parameters:
          queue_name(str): The queue to consume.
          prefetch(int): The number of messages to prefetch.
          timeout(int): The idle timeout in milliseconds.

        Raises:
          QueueNotFound: If the queue hasn't been declared.

        Returns:
          Consumer: A consumer that retrieves messages from Redis.
        """
        try:
            return _StubConsumer(self.queues[queue_name], self.dead_letters, timeout)
        except KeyError:
            raise QueueNotFound(queue_name)

    def declare_queue(self, queue_name):
        """Declare a queue.  Has no effect if a queue with the given
        name has already been declared.

        Parameters:
          queue_name(str): The name of the new queue.
        """
        if queue_name not in self.queues:
            self.emit_before("declare_queue", queue_name)
            self.queues[queue_name] = Queue()
            self.emit_after("declare_queue", queue_name)

            delayed_name = dq_name(queue_name)
            self.queues[delayed_name] = Queue()
            self.delay_queues.add(delayed_name)
            self.emit_after("declare_delay_queue", delayed_name)

    def enqueue(self, message, *, delay=None):
        """Enqueue a message.

        Parameters:
          message(Message): The message to enqueue.
          delay(int): The minimum amount of time, in milliseconds, to
            delay the message by.

        Raises:
          QueueNotFound: If the queue the message is being enqueued on
            doesn't exist.
        """
        queue_name = message.queue_name
        if delay is not None:
            queue_name = dq_name(queue_name)
            message_eta = current_millis() + delay
            message = message.copy(queue_name=queue_name, options={"eta": message_eta})

        self.emit_before("enqueue", message, delay)

        if queue_name not in self.queues:
            raise QueueNotFound(queue_name)

        self.queues[queue_name].put(message.encode())
        self.emit_after("enqueue", message, delay)
        return message

    def flush(self, queue_name):
        """Drop all the messages from a queue.

        Parameters:
          queue_name(str): The queue to flush.
        """
        for _ in iter_queue(self.queues[queue_name]):
            self.queues[queue_name].task_done()

    def flush_all(self):
        """Drop all messages from all declared queues."""
        for queue_name in chain(self.queues, self.delay_queues):
            self.flush(queue_name)

    def join(self, queue_name, *, timeout=None):
        """Wait for all the messages on the given queue to be
        processed.  This method is only meant to be used in tests
        to wait for all the messages in a queue to be processed.

        Raises:
          QueueJoinTimeout: When the timeout elapses.
          QueueNotFound: If the given queue was never declared.

        Parameters:
          queue_name(str): The queue to wait on.
          timeout(Optional[int]): The max amount of time, in
            milliseconds, to wait on this queue.
        """
        try:
            deadline = timeout and time.monotonic() + timeout / 1000
            while True:
                for name in [queue_name, dq_name(queue_name)]:
                    timeout = deadline and deadline - time.monotonic()
                    join_queue(self.queues[name], timeout=timeout)

                # We cycle through $queue then $queue.DQ then $queue
                # again in case the messages that were on the DQ got
                # moved back on $queue.
                for name in [queue_name, dq_name(queue_name)]:
                    if self.queues[name].unfinished_tasks:
                        break
                else:
                    return
        except KeyError:
            raise QueueNotFound(queue_name)


class _StubConsumer(Consumer):
    def __init__(self, queue, dead_letters, timeout):
        self.queue = queue
        self.dead_letters = dead_letters
        self.timeout = timeout

    def ack(self, message):
        self.queue.task_done()

    def nack(self, message):
        self.queue.task_done()
        self.dead_letters.append(message)

    def __next__(self):
        try:
            data = self.queue.get(timeout=self.timeout / 1000)
            message = Message.decode(data)
            return MessageProxy(message)
        except Empty:
            return None
